/* GUI.java
 * Graphical SLIME interface, using AWT.
 *
 * Version 0.1 - Created 9/12/96 by Mark Gritter 
 *
 * Version 0.2 - Modified 9/19/96 by Mark Gritter
 *   Added error message for bogus PC.
 *   Invalid number entry now leaves value unchanged.
 *   Changed layout so that registers are no longer in a GridLayout; i.e., 
 *   they now remain the same size when the window is resized.
 *   "Load" and "Reset" options now wait for Machine thread to actually 
 *   stop before mucking with the machine state.
 *   The "Stop" and "Reset" buttons cancel console input.
 *   The input of the console is cleared after being read.
 *   Added indicator to reflect/change exception flag.
 *   Added instruction counter and clear button.
 *
 * Version 0.21 - Modified 9/22/96 by Mark Gritter
 *   Need to get both file and directory from FileDialog.
 *
 * Version 0.22 - Modified 3/4/97 by Max Hailperin <max@gac.edu>
 *   Made error dialog a class of its own so as to make OK button work.
 *
 * Version 0.3 - Modified 7/28/97 by Mark Gritter
 *   Added "Quit" button.
 *   Changed instruction and data windows to use Courier-plain-12 font.
 *   Added "dispose" method to clean up console window.
 *
 * Version 0.4 - Modified 7/30/97 by Max Hailperin <max@gac.edu>
 *   Made load command also select and scroll to the PC instruction;
 *   made selecting an instruction with mouse change the PC.
 */

package EDU.gac.mgritter.SLIME;

import java.lang.*;
import java.awt.*;
import java.io.*;

/**
 * GUI for a single simulated machine.  Displays registers, program text, 
 * and memory.  Has buttons for start and stop.  PC field is editable.
 */

public class GUI extends Frame implements UserInterface {
  private TextField PC, icounter;
  private TextField[] register;
  private Button start, stop, step, reset, load, quit, clearcounter;
  private List instructions, data;
  private Label status;
  private Console console;
  private Dialog error_dialog;
  private TextArea error_text;
  private Indicator indicat;

  protected MachineThread thread;

  public GUI(MachineThread t) {
    this(t, true);
  }

  public GUI(MachineThread t, boolean showconsole) {
    super("SLIME " + SLIME.version);
    thread = t;
    
    Panel p, p2, p3, large;

    p = new Panel();
    p.setLayout(new GridLayout(2, 1));

    p2 = new Panel();
    // Program Counter
    p2.add(new Label("PC"));
    PC = new TextField("0", 10);
    p2.add(PC);

    start = new Button("Start");
    stop = new Button("Stop");
    step = new Button("Step");
    reset = new Button("Reset");
    load = new Button("Load");
    quit = new Button("Quit");
    p2.add(start);
    p2.add(stop);
    p2.add(step);
    p2.add(reset);
    p2.add(load);
    p2.add(quit);
    p.add(p2);

    // Instruction counter
    p2 = new Panel();
    p2.add(new Label("Counter"));
    icounter = new TextField("0", 10);
    icounter.setEditable(false);
    p2.add(icounter);
    clearcounter = new Button("Clear Counter");
    p2.add(clearcounter);
    p.add(p2);
    add("North", p);
    
    // "large" contains a (pretty-much) fixed register panel and 
    // a grid for the data and instructions
    large = new Panel();
    large.setLayout(new BorderLayout());

    // Registers
    p2 = new Panel();
    p2.setLayout(new BorderLayout());
    p2.add("North", new Label("Registers", Label.CENTER));
    register = new TextField[32];
    p = new Panel();
    p.setLayout(new GridLayout(4, 8));
    int i;
    for (i = 0; i < 32; i++) {
	register[i] = new TextField("0", 10);
	p3 = new Panel();
	p3.setLayout(new BorderLayout());
	p3.add("West", new Label("R" + (i < 10 ? "0" : "") + i, Label.RIGHT));
	p3.add("Center", register[i]);
	p.add(p3);
    }
    p2.add("Center", p);
    large.add("North", p2);

    p3 = new Panel();
    p3.setLayout(new GridLayout(2, 1));

    Font courier = new Font("Courier", Font.PLAIN, 12);
    // Program
    p2 = new Panel();
    p2.setLayout(new BorderLayout());
    p2.add("North", new Label("Instructions", Label.CENTER));
    instructions = new List();
    if (courier != null) instructions.setFont(courier);
    p2.add("Center", instructions);
    p3.add(p2);
    
    
    // Data
    p2 = new Panel();
    p2.setLayout(new BorderLayout());
    p2.add("North", new Label("Data", Label.CENTER));
    data = new List();
    if (courier != null) data.setFont(courier);
    p2.add("Center", data);
    p3.add(p2);
    large.add("Center", p3);
    

    add("Center", large);

    // Status bar
    p = new Panel();
    p.setLayout(new BorderLayout());
    indicat = new Indicator("EXC", "OK", "Exception");
    p.add("West", indicat);
    status = new Label("", Label.LEFT);
    p.add("Center", status);
    add("South", p);

    resize(800, 600);
    show();

    // Error dialog box.
    error_dialog = new ErrorDialog(this, "Parsing Errors", false); 
    error_text = new TextArea();
    error_dialog.add("Center", error_text);
    error_dialog.add("South", new Button("OK"));
    error_dialog.resize(600,400);
 
    // Console
    if (showconsole) {
      console = new Console();
      console.resize(400,300);
      console.show();
    } else {
      console = null;
    }
  }

  private String createLine(MachineState s, int pos, int max) {
    String wk;
    if (pos == 0) {
      wk = " 0- 9: ";
    } else {
      wk = String.valueOf(pos) + "-" + String.valueOf(pos + 9) + ": ";
    }
    int i;
    for (i = 0; i < max; i++, pos++) {
     wk += (i == 0 ? "" : ", ") + s.getWord(pos);
    }
    return wk;
  }

  public void dataChanged(int address, int newvalue) {
    int line = address / 10;
    int pos = address % 10;
    MachineState state = thread.getState();
    while (data.countItems() <= line) {
      int l = data.countItems();
      data.addItem(createLine(state, l * 10, 10), l);
    }
    data.replaceItem(createLine(state, line * 10, 10), line);
  }

  public void instructionChanged(int address, Instruction newinst) {
    while (address >= instructions.countItems()) {
      instructions.addItem("***");
    }
    String txt;
    if (newinst == null) {
      txt = "";
    } else {
      txt = newinst.toString();
    }
    instructions.replaceItem((address < 10 ? " " : "") + 
			     address + ": " + txt, address);
  }

  public void registerChanged(int reg, int newvalue) {
    register[reg].setText(String.valueOf(newvalue));
  }
  
  public void PCChanged(int newpc) {
    PC.setText(String.valueOf(newpc));
    instructions.select(newpc);
    instructions.makeVisible(newpc);
  }

  public void exceptionRaised(String message) {
    status.setText(message);
    switch (thread.getState().getCause()) {
    case MachineState.EX_NONE:
    case MachineState.EX_INPINT:
      indicat.setState(false);
      break;
    default:
      indicat.setState(true);
      break;
    }
  }

  public boolean action(Event evt, Object o) {
    Object target = evt.target;
    if (target instanceof Button) {
      int cause = thread.getState().getCause();
      if (o.equals("Start")) {
        if (cause == MachineState.EX_BADINPUT) {
	  thread.getState().clearException();
	}
	thread.startExecution();
	return true;
      } else if (o.equals("Stop")) {
	if (console != null) console.cancelInput();
	thread.stopExecution();
	return true;
      } else if (o.equals("Step")) {
	thread.stepExecution();
	return true;
      } else if (o.equals("Reset")) {
	if (console != null) console.cancelInput();
	thread.stopAndWait();	
	thread.getState().reset();
	thread.clearCount();
	return true;
      } else if (o.equals("Load")) {
	if (console != null) console.cancelInput();
	thread.stopAndWait();
	if (SLIME.isApplet()) {
	  thread.getState().loadFromString(SLIME.getApplet().getProgram());
	} else {
	  FileDialog fd = new FileDialog(this, "Load SLIM Program");
	  String filename;
	  fd.resize(400,500);
	  fd.show();
	  filename = fd.getFile();
	  if (filename != null) {
	    filename = fd.getDirectory() + filename;
	    thread.getState().loadFromFile(filename);
	  }
	}
        try {
	  int pc = Integer.parseInt(PC.getText());
          instructions.select(pc);
          instructions.makeVisible(pc);
	}
	catch (NumberFormatException e) {}
	return true;
      } else if (o.equals("Quit")) {
	thread.stopExecution();
	if (!SLIME.isApplet()) {
	  System.exit(0);
	} else {
	  SLIME.getApplet().quit();
	}
	return true;
      } else if (o.equals("OK")) {
	// This message is received when the ErrorDialog's OK button 
        // gets clicked.
	error_dialog.hide();
	error_text.setText("");
	return true;
      } else if (o.equals("Clear Counter")) {
	thread.clearCount();
	return true;
      }
    }

    if (target instanceof Indicator && o.equals("EXC")) {
      // Only do this if control has been set-- i.e., program 
      // already stopped.
      if (indicat.state()) {
	thread.getState().clearException();
      }
      return true;
    }

    if (target instanceof TextField) {
      int r, v;
      for (r = 0; r < 32; r++) {
	if (target.equals(register[r])) {
	  try {
	    v = Integer.parseInt(register[r].getText());
	    status.setText("");
	    thread.getState().setRegister(r, v);
	  }
	  catch (NumberFormatException e) {
	    status.setText("Bad number format for Register.");
	  }
	  return true;
	}
      }
      if (target.equals(PC)) {
	try {
	  v = Integer.parseInt(PC.getText());
	  thread.getState().setPC(v);
	  thread.getState().clearException();
	}
	catch (NumberFormatException e2) {
	  status.setText("Bad number format for PC.");
	}
	return true;
      }
    }
    return false;
  }

  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY) {
      System.exit(0);
    } else if(evt.id == Event.LIST_SELECT && evt.target == instructions){
      int newpc = ((Integer) evt.arg).intValue();
      PC.setText(String.valueOf(newpc));
      thread.getState().setPC(newpc);
      thread.getState().clearException();
      return true;
    } else if(evt.id == Event.LIST_DESELECT && evt.target == instructions){
      int pc = ((Integer) evt.arg).intValue();
      instructions.select(pc); // they deselect, we reselect = can't deselect
      return true;
    }
    return super.handleEvent(evt);
  }

  /**
   * Read integer from console.
   */
  public int readConsole() {
    if (console == null) {
      return -1;
    } else {
      String buf = console.getLine();
      // If input has been canceled, return 0,
      // and make sure PC doesn't change
      if (buf.equals("")) {
	thread.getState().raiseException(MachineState.EX_INPINT);
	return 0;
      }
      try {
	return Integer.parseInt(buf);
      }
      catch (NumberFormatException e) {
        thread.getState().raiseException(MachineState.EX_BADINPUT);
	return 0;
      }
    }
  }
  
  /**
   * Write integer to console.
   */
  public void writeConsole(int n) {
    if (console != null) {
      console.write(String.valueOf(n));
    }
  }

  /**
   * Show a list of errors as a dialog box.
   */
  public void showErrors(String[] errors) {
    int i;
    error_text.setText("");
    for (i = 0; i < errors.length; i++) {
      error_text.appendText(errors[i] + "\n");
    }
    error_dialog.show();
  }

  /**
   * Show the instruction count.
   */
  public void countChanged(int count) {
    icounter.setText(String.valueOf(count));
  }

  /** 
   * Close the console window.
   */
  public void dispose() {
    if (console != null) {
      console.dispose();
    }
    super.dispose();
  }
}

class Console extends Frame {
  private TextArea area;
  private String buffer;
  private Button clear;
  private TextField input;
  private boolean cancel, waiting;
 
  public Console() {
    super("SLIM Console");

    area = new TextArea();
    area.setEditable(false);
    add("Center", area);
    
    Panel p = new Panel();
    p.setLayout(new GridLayout(2, 1));
    input = new TextField();
    p.add(input);
    clear = new Button("Clear Console");
    p.add(clear);

    add("South", p);
    input.disable();
    waiting = false;
  }

  public synchronized void cancelInput() {
    if (waiting) {
      cancel = true;
      write("");
      notify();
    }
  }

  public synchronized String getLine() {
    // Show prompt, make sure window is showing, set flags
    setTitle("Console: waiting for input...");
    show();
    buffer = "";
    input.enable();
    input.requestFocus();
    area.appendText(">>> ");
    cancel = false;
    waiting = true;
    // Wait on UI thread for user input
    while (buffer.length() == 0 && cancel == false) {
      try {
	wait();
      }
      catch (InterruptedException ie) {
      }
    }
    // Restore original status before clearing wait flag
    setTitle("SLIM Console");
    input.disable();
    input.setText("");
    waiting = false;

    return buffer;
  }

  public boolean isActive() {
    return waiting;
  }

  public synchronized void write(String txt) {
    area.appendText(txt + "\n");
  }

  public synchronized void putLine(String line) {
    if (line.length() > 0) {
      write(line);
      buffer = line;
      notify();
    }
  }

  public synchronized void clear() {
    area.setText("");
  }

  public boolean action(Event evt, Object o) {
    if (o.equals("Clear Console")) {
      clear();
      return true;
    } else if (evt.target == input) {
      putLine((String)o);
      return true;
    }
    return false;
  }
}

class ErrorDialog extends Dialog {
  private Frame parent;

  ErrorDialog(Frame p, String s, boolean b){
    super(p, s, b);
    parent = p;
  }

  public boolean action(Event evt, Object o) {
    if (evt.target instanceof Button && o.equals("OK")) {
      // The "GUI" class will handle closing down the dialog.
      return parent.action(evt, o);
    }
    return false;
  }

  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY) {
      this.hide();
      return true;
    }
    return super.handleEvent(evt);
  }
}

