/* MachineState.java
 * Simulated registers, PC, instruction and data memory, and exceptions.
 *
 * Version 0.1 -- Created 9/12/96 by Mark Gritter 
 *
 * Version 0.2 -- Modified 9/19/96 by Mark Gritter
 *   Fixed jump to PC=0 bug, changed > to >= in MachineState.jump
 *   Synchronized "setPC" and "setRegister", since they are accessed 
 *    by both the Machine and GUI threads.
 *
 * Version 0.3 -- Modified 7/28/97 by Mark Gritter
 *   Created two version of "load" function-- one for file input, 
 *    one for reading directly from a String.
 *
 * Version 0.4 -- Modified 7/31/97 by Max Hailperin <max@gac.edu>
 *   Added clearInstructionsFrom method to allow the instruction memory
 *    to be purged when re-loaded, so that users wouldn't have the debugging
 *    hassles of tail ends of old, longer programs tacked onto new, shorter
 *    programs.
 */

package EDU.gac.mgritter.SLIME;

import java.lang.*;
import java.io.*;

/**
 * A single machine state.
 */
public class MachineState {
  private InstructionMemory instructions;
  private Register[] registers;
  private int PC;
  private boolean exception;
  private int ex_cause;
  private int[] data;
  private boolean jumped;

  private UserInterface ui;

  public final static int NUMREGS = 32;

  public static final int EX_NONE = 0;
  public static final int EX_DIVBYZERO = 1;
  public static final int EX_OVERFLOW = 2;
  public static final int EX_NOINST = 3;
  public static final int EX_STOP = 4;
  public static final int EX_BADINPUT = 5;
  public static final int EX_NODATA = 6;
  public static final int EX_INPINT = 7;

  /**
   * Create a new machine state, with empty memory.
   *
   * @param userinterface An object implementing UserInterface, to use 
   * for communicating with the user.
   */
  public MachineState(UserInterface userinterface) {
    instructions = new InstructionMemory();
    registers = new Register[NUMREGS];
    int i;
    for (i = 0; i < NUMREGS; i++) registers[i] = new Register();
    exception = false;
    ex_cause = 0;
    PC = 0;
    data = new int[10];
    ui = userinterface;
    jumped = false;
  }

  /**
   * Reset machine by zeroing registers, memory(?), and PC.
   * Clear any exception.
   */
  public void reset() {
    int i;
    clearException();
    for (i = 0; i < NUMREGS; i++) setRegister(i, 0);
    for (i = 0; i < data.length; i++) setWord(i, 0);
    setPC(0);
    
  }

  /**
   * Read a machine register.
   */
  public int getRegister(int reg) {
    return registers[reg].get();
  }

  /**
   * Set a machine register.
   */
  public synchronized void setRegister(int reg, int value) {
    registers[reg].set(value);
    ui.registerChanged(reg, value);
  }

  /**
   * Read the program counter.
   */
  public int getPC() {
    return PC;
  }

  /**
   * Return true if a jump has occurred, reset flag.
   */
  public boolean jumpOccurred() {
    if (jumped) {
      jumped = false;
      return true;
    }
    return false;
  }

  /** 
   * Jump to a new location.
   */
  public void jump(int address) {
    if (address >= 0) {
      jumped = true;
      setPC(address);
    } else {
      raiseException(EX_NOINST);
    }
  }

  /**
   * Set the program counter.
   */
  public synchronized void setPC(int pc) {
    PC = pc;
    ui.PCChanged(pc);
  }

  /**
   * Get an instruction from memory.
   */
  public Instruction getInstruction(int address) {
    try {
      return instructions.get(address);
    }
    catch (ArrayIndexOutOfBoundsException e) {
      raiseException(EX_NOINST);
      return null;
    }
  }

  /**
   * Set an instruction in memory.
   */
  public void setInstruction(int address, Instruction i) {
    instructions.set(address, i);
    ui.instructionChanged(address, i);
  }
  
  /**
   * Clear all instructions in memory from given address onward.
   */
  public void clearInstructionsFrom(int address) {
    for( ; address < instructions.size(); address++){
      setInstruction(address, null);
    }
  }
  
  /**
   * Expand memory.
   */
  public void expandDataMemory(int minaddr) {
    if (minaddr >= data.length) {
      int newsize = (minaddr / 10 + 1) * 10;
      int[] newdata = new int[newsize];
      int i;
      for (i = 0; i < data.length; i++) {
	newdata[i] = data[i];
      }
      for (i = data.length; i < newsize; i++) {
	newdata[i] = 0;
      }
      data = newdata;
    }
  }

  /**
   * Get a word from data memory.
   */
  public int getWord(int address) {
    expandDataMemory(address);
    return data[address];
  }

  /**
   * Set a word in data memory.
   */
  public void setWord(int address, int word) {
    expandDataMemory(address);
    data[address] = word;
    ui.dataChanged(address, word);
  }
 
  /**
   * Raise an exception.
   */
  public void raiseException(int cause) {
    String msg;
    exception = true;
    ex_cause = cause;
    switch (cause) {
    case EX_NONE: msg = ""; exception = false; break;
    case EX_DIVBYZERO: msg = "Division by zero at instruction " + PC + "."; break;
    case EX_OVERFLOW: msg = "Overflow at instruction " + PC + "."; break;
    case EX_NOINST: msg = "No instruction found at " + PC + "."; break;
    case EX_STOP: msg = "Program stoppped at " + PC + "."; break;
    case EX_BADINPUT: msg = "Bad input at " + PC + "."; break;
    case EX_NODATA: 
      msg = "Access past end of data memory at instruction " + PC + "."; 
      break;
    case EX_INPINT: msg = ""; break;
    default: msg = "Unidentified exception at " + PC + "?!"; break;
    }
    ui.exceptionRaised(msg);
  }

  /**
   * Return exception cause, or 0 for none.
   */
  public int getCause() {
    return ex_cause;
  }

  /**
   * Clear all exceptions.
   */
  public void clearException() {
    raiseException(EX_NONE);
  }

  /**
   * Read integer from console.
   */
  public int readConsole() {
    return ui.readConsole();
  }
  
  /**
   * Write integer to console.
   */
  public void writeConsole(int n) {
    ui.writeConsole(n);
  }

  /**
   * Load and parse a file.
   */
  public void loadFromFile(String filename) {
    try {
      FileInputStream fis = new FileInputStream(filename);
      SLIMParser parser = new SLIMParser(fis, this);
      String[] errors = parser.parse();
      fis.close();
      if (errors.length > 0) ui.showErrors(errors);
      ui.exceptionRaised(filename + " loaded.");
    }
    catch (IOException e) {
      ui.exceptionRaised("Error opening " + filename + ".");
    }
  }

  /**
   * Load a string as a program.
   */
  public void loadFromString(String s) {
    try {
      StringBufferInputStream sbis = new StringBufferInputStream(s);
      SLIMParser parser = new SLIMParser(sbis, this);
      String[] errors = parser.parse();
      if (errors.length > 0) ui.showErrors(errors);
      ui.exceptionRaised("Program loaded.");
    }
    catch (IOException e) {
      ui.exceptionRaised("Error loading program.");
    }
  }
}







