/* MachineThread.java
 * Start, stop, and run the simulator.
 *
 * Version 0.1 - Created 9/12/96 by Mark Gritter
 *
 * Version 0.2 - Modified 9/19/96 by Mark Gritter
 *   Added synchronized keywords and new noStep and nextInstruction 
 *   functions.
 *   Changed busy waiting in run() to wait()/notify().
 *   Added function stop_and_wait.
 *   Added instruction counter.
 *
 * Version 0.3 - Modified 7/28/97 by Mark Gritter
 *   Added "stopMachine" method to allow User Interface dispose()al.
 */

package EDU.gac.mgritter.SLIME;

import java.lang.*;

/**
 * Thread to run simulation of SLIM machine.  Creates a machine state and
 * user interface.
 */

public class MachineThread extends Thread {
  private MachineState state;
  private UserInterface user;
  private boolean stepping, running, waiting;
  private int count;

  public MachineState getState() {
    return state;
  }

  MachineThread() {
    user = new GUI(this);
    state = new MachineState(user);
    stepping = false;
    running = false;
  }

  public synchronized void stepExecution() {
    if (!stepping) {
      stepping = true;
      running = true;
      notify();
    }
  }

  public synchronized void stopExecution() {
    running = false;
    stepping = false;
  }
    
  public synchronized void startExecution() {
    if (!running) {
      stepping = false;
      running = true;
      notify();
    }
  }

  private synchronized void noStep() {
    running = false;
    stepping = false;
  }


  /**
   * Wait for the "running" flag.
   */
  private synchronized void nextInstruction() {
    waiting = true;
    // Anybody who's waiting for us to stop can go now.
    notify();
    while (!running) { 
      try {
	wait();
      }
      catch (InterruptedException e) {
	// ???
      }
    }
    waiting = false;
  }

  /**
   * Halt the simulator and do not return until the machine thread
   * has begun waiting.
   */
  public synchronized void stopAndWait() {
    stopExecution();
    while (!waiting) {
      try {
	wait();
      }
      catch (InterruptedException e) {
      }
    }
  }

  public void run() {
    int pc;
    Instruction i;
    while (true) {
      nextInstruction(); 
      pc = state.getPC();
      i = state.getInstruction(pc);
      if (i == null) {
	state.raiseException(MachineState.EX_NOINST);
      } else {
	i.perform(state);
	if (!state.jumpOccurred()) state.setPC(pc + 1);
      }
      // If an exception thrown, stop and restore old PC.
      if (state.getCause() != 0) {
	running = false;
	state.setPC(pc);
	// The INPINT exception occurs only to make sure the PC
	// gets reset correctly-- then it should disappear.
	if (state.getCause() == MachineState.EX_INPINT) {
	  state.clearException();
	}
	// The EX_STOP exception needs to be counted as an 
	// excuted instruction, even though PC doesn't change.
	if (state.getCause() == MachineState.EX_STOP) {
	  count++;
	  user.countChanged(count);
	}
      } else {
	// Instruction successful.
	count++;
	user.countChanged(count);
      }
      // Only step once.
      if (stepping) noStep();
    }
  }

  public void clearCount() {
    count = 0;
    user.countChanged(count);
  }

  public void stopMachine() {
    user.dispose();
    stop();
  }
}




