/* MathInstruction.java
 * Created 9/12/96 by Mark Gritter 
 */

package EDU.gac.mgritter.SLIME;

import java.lang.*;

/**
 * A mathematical operation (including comparisons.)  
 * Includes three operands: a destination register and two 
 * source registers.
 */
public class MathInstruction extends Instruction {
  protected int destreg;
  protected int sourcereg1, sourcereg2;

  private int op;
  public static final int ADD = 1, SUB = 2, MUL = 3, DIV = 5, QUO = 6, REM = 7;
  public static final int SEQ = 8, SNE = 9, SLT = 10, SGT = 11, SLE = 12, SGE = 13;

  /**
   * Create a new instruction.
   *
   * @param operation One of ADD, SUB, MUL, DIV, QUO, REM, SEQ, SNE, SLT, SGT,
   * SLE, or SGE.
   * @param dest Destination register.
   * @param source1 Source register 1.
   * @param source2 Source register 2.
   */
  public MathInstruction(int operation, int dest, int source1, int source2) {
    op = operation;
    destreg = dest;
    sourcereg1 = source1;
    sourcereg2 = source2;
  }
  
  public int getDestReg() { return destreg; }
  public int getSourceReg1() { return sourcereg1; }
  public int getSourceReg2() { return sourcereg2; }

  public String mnemonic() {
    switch (op) {
    case ADD: return "add";
    case SUB: return "sub";
    case MUL: return "mul";
    case DIV: return "div";
    case QUO: return "quo";
    case REM: return "rem";
    case SEQ: return "seq";
    case SNE: return "sne";
    case SLT: return "slt";
    case SGT: return "sgt";
    case SLE: return "sle";
    case SGE: return "sge";
    }
    return "MAT?";
  }

  public String arguments() {
    return formatReg(destreg) + ", " + 
           formatReg(sourcereg1) + ", " + 
           formatReg(sourcereg2);
  }

  public void perform(MachineState state) {
    int s1 = state.getRegister(sourcereg1);
    int s2 = state.getRegister(sourcereg2);
    int result = state.getRegister(destreg);
    boolean conditional = false, temp = false;

    /* Preprocessing */
    switch (op) {
    case DIV:
    case QUO:
    case REM:
      if (s2 == 0) {
	state.raiseException(MachineState.EX_DIVBYZERO);
	return;
      }
      break;
    case SEQ:
    case SNE:
    case SLT:
    case SGT:
    case SLE:
    case SGE:
      conditional = true;
      break;
    }

    /* Calculation */
    switch (op) {
    case ADD: result = s1 + s2; break;
    case SUB: result = s1 - s2; break;
    case MUL: result = s1 * s2; break;
    case DIV:
    case QUO: result = s1 / s2; break;
    case REM: result = s1 % s2; break;
    case SEQ: temp = (s1 == s2); break;
    case SNE: temp = (s1 != s2); break;
    case SLT: temp = (s1 < s2); break;
    case SGT: temp = (s1 > s2); break;
    case SLE: temp = (s1 <= s2); break;
    case SGE: temp = (s1 >= s2); break;
    }
    
    if (conditional) {
      result = (temp ? 1 : 0);
    }
    state.setRegister(destreg, result);
  }
}






