/* SLIMParser.java 
 * Parsing code for SLIM assembly language (more or less).
 *
 * Version 0.1 -- Created 9/16/96 by Mark Gritter 
 * 
 * Version 0.2 -- Modified 9/19/96 by Mark Gritter
 *  In SLIMParser.parse:
 *   Changed register allocation pool to start with 0 rather than 1.
 *   Changed search for "allocate-registers" keyword.
 *   Line number variable "i" ranges from 1 to length, not 0 to length-1.
 *    Thus, error messages don't have to add 1 any more.
 *   Created new const string to hold delimiters.
 *   StringTokenizer used to determine if line is empty now uses 
 *    "\t ," instead of " :"  for delimiters.
 *  In SLIMParser.parseReg:
 *   Rejects negative register numbers.
 *  In SLIMParser.parseWord:
 *   Allows negative numbers.
 *
 * Version 0.3 -- Modified 7/28/97 by Mark Gritter
 *   Labels are hashed twice: once by name, once by address.
 *    This allows instructions to have (one) label associated with them.
 *   The constructor takes an InputStream instead of a filename.
 *
 * Version 0.4 -- Modified 7/30/97-7/31/97 by Max Hailperin <max@gac.edu>
 *   Provided error messages for use of label as register name or vice versa
 *   and for re-declaration of a label or register name.
 */

package EDU.gac.mgritter.SLIME;

import java.lang.*;
import java.io.*;
import java.util.*;

/**
 * The SLIMParser class takes an input stream and a machine state, and 
 * parses the stream into instruction memory.  Currently, memory is 
 * cleared and the file is loaded starting at 0.  This may change 
 * in the future.
 */
public class SLIMParser {
  private Hashtable labels, labelsbyaddr, labelTypes;
  private Hashtable mathop, ioop, specop, jumpop, immop, condop, memop;
  private Vector lines;
  private Vector errors;
  private MachineState state;

  /**
   * Delimiters for opcodes, arguments -- unimportant "white space."
   */
  private final static String delim = "\t ,";

  /**
   * Load the program from an input stream.
   */
  public SLIMParser(InputStream is, MachineState s) throws IOException {
    state = s;
    DataInputStream in = new DataInputStream(is);
    labels = new Hashtable();
    labelsbyaddr = new Hashtable();
    labelTypes = new Hashtable();
    lines = new Vector();
    String l;
    while ((l = in.readLine()) != null) {
      lines.addElement(l);
    }
    in.close();
    mathop = new Hashtable();
    mathop.put("add", new Integer(MathInstruction.ADD));
    mathop.put("sub", new Integer(MathInstruction.SUB));
    mathop.put("mul", new Integer(MathInstruction.MUL));
    mathop.put("div", new Integer(MathInstruction.DIV));
    mathop.put("quo", new Integer(MathInstruction.QUO));
    mathop.put("rem", new Integer(MathInstruction.REM));
    mathop.put("seq", new Integer(MathInstruction.SEQ));
    mathop.put("sne", new Integer(MathInstruction.SNE));
    mathop.put("slt", new Integer(MathInstruction.SLT));
    mathop.put("sgt", new Integer(MathInstruction.SGT));
    mathop.put("sle", new Integer(MathInstruction.SLE));
    mathop.put("sge", new Integer(MathInstruction.SGE));
    ioop = new Hashtable();
    ioop.put("read", new Integer(IOInstruction.READ));
    ioop.put("write", new Integer(IOInstruction.WRITE));
    specop = new Hashtable();
    specop.put("stop", new Integer(SpecialInstruction.STOP));
    specop.put("halt", new Integer(SpecialInstruction.STOP));
    jumpop = new Hashtable();
    jumpop.put("j", new Integer(JumpInstruction.J));
    immop = new Hashtable();
    immop.put("li", new Integer(ImmediateInstruction.LI));
    condop = new Hashtable();
    condop.put("jeqz", new Integer(ConditionalInstruction.JEQZ));
    memop = new Hashtable();
    memop.put("st", new Integer(MemoryInstruction.ST));
    memop.put("ld", new Integer(MemoryInstruction.LD));
  }

  /**
   * Parse the file, return error string(s). 
   */
  public String[] parse() {
    String wk;
    StringTokenizer st;
    int i, pos, n;
    int a = 0;
    int available_reg = 0;

    errors = new Vector();
 
    /* First pass
     * Remove comments.  
     * Find and remove labels (before colons) 
     * Handle register allocation, marked by "allocate-registers" line */
    for (i = 1; i <= lines.size(); i++) {
      wk = (String)lines.elementAt(i - 1);
      try {
	// Remove comment
	pos = wk.indexOf(";");
	if (pos != -1) {
	  wk = wk.substring(0, pos);
	}
	// Check for allocate-registers
	pos = wk.indexOf("allocate-registers ");
	if (pos != -1) {
	  // Make sure the keyword isn't part of something wierd.
	  if (pos != 0) {
	    String check = wk.substring(0, pos);
	    int ch;
	    for (ch = 0; ch < pos; ch++) {
	      if (!Character.isSpace(check.charAt(ch))) {
		errors.addElement("Bad use of 'allocate-registers' keyword on line " + i);
		continue;
	      }
	    }
	  }
	  wk = wk.substring(pos + 19);
	  st = new StringTokenizer(wk, delim);
	  while (st.hasMoreTokens()) {
	    String name = st.nextToken();
            if(labels.get(name) != null){
              errors.addElement("Register name '" + name +
                                "' already in use on line " + i + ".");
	    } else if (available_reg == MachineState.NUMREGS) {
	      errors.addElement("No more registers available for '" + name + "'.");
	    } else {
	      labels.put(name, new Integer(available_reg++));
              labelTypes.put(name, "register name");
	    }
	  }
	  wk = "";
	}
	// Check for line label
        st = new StringTokenizer(wk, ":");
	n = st.countTokens();
	if (n > 2) {
	  errors.addElement("Too many colons in line " + i + ".");
	} else if (n == 2 || (n == 1 && wk.indexOf(":") != -1)) {
	  String name = st.nextToken(": \t");
	  String rest;
	  if (n == 1) 
	    rest = "";
	  else 
	    rest = st.nextToken(":");
	  // System.err.println(name + " = " + a);
          if(labels.get(name) != null){
            errors.addElement("Label '" + name +
                                "' already in use on line " + i + ".");
          } else {
            labels.put(name, new Integer(a));
            labelsbyaddr.put(new Integer(a), name);
            labelTypes.put(name, "label");
          }
	  wk = rest;
	}
	// Strip off any leading whitespace.
	wk = wk.trim();

	// System.err.println(a + ": " + wk);

	// If this line isn't blank after all the chopping we've done,
	// it must have an instruction.
	if ((new StringTokenizer(wk, delim)).countTokens() != 0) a++;

	lines.setElementAt(wk, i - 1);
      }
      catch (Exception e1) {
	System.err.println("Line " + i + "('" + wk + "'): " + e1);
	errors.addElement("Internal problem, see error log.");
      }
    }

    a = 0;
    if (errors.size() == 0) {
      /* Second pass - find opcodes */
      for (i = 1; i <= lines.size(); i++) {
	wk = (String)lines.elementAt(i - 1);
	st = new StringTokenizer(wk, delim);
	if (st.countTokens() != 0) {
	  Instruction inst;
	  try {
	    String opcode = st.nextToken().toLowerCase();
	    Integer op;
	    if ((op = (Integer)mathop.get(opcode)) != null) {
	      inst = new MathInstruction(op.intValue(), parseReg(st, i), 
					 parseReg(st, i), parseReg(st, i));
	    } else if ((op = (Integer)ioop.get(opcode)) != null) {
	      int ioop_reg, ioop_port = 0;
	      ioop_reg = parseReg(st, i);
	      if (st.hasMoreTokens()) {
		ioop_port = parseWord(st, i);
	      }
	      inst = new IOInstruction(op.intValue(), ioop_reg, ioop_port);
	    } else if ((op = (Integer)specop.get(opcode)) != null) {
	      inst = new SpecialInstruction(op.intValue());
	    } else if ((op = (Integer)jumpop.get(opcode)) != null) {
	      inst = new JumpInstruction(op.intValue(), parseReg(st, i));
	    } else if ((op = (Integer)immop.get(opcode)) != null) {
	      inst = new ImmediateInstruction(op.intValue(), parseReg(st, i), 
					      parseWord(st, i));
	    } else if ((op = (Integer)condop.get(opcode)) != null) {
	      inst = new ConditionalInstruction(op.intValue(), 
						parseReg(st, i), 
						parseReg(st, i));
	    } else if ((op = (Integer)memop.get(opcode)) != null) {
	      inst = new MemoryInstruction(op.intValue(), 
					   parseReg(st, i),
					   parseReg(st, i));
	    } else {
	      errors.addElement("Unknown opcode '" + opcode + "' in line " + i + ".");
	      inst = null;
	    }
	    if (inst != null && st.hasMoreTokens()) {
	      inst = null;
	      errors.addElement("Too many arguments in line " + i + ".");
	    }
	    if (inst != null) {
	      inst.setText(wk);
	      String instlabel = (String)labelsbyaddr.get(new Integer(a));
	      if (instlabel != null) {
		inst.setLabel(instlabel);
	      }
	    }
	  }
	  catch (NoSuchElementException e2) {
	    errors.addElement("Missing argument in line " + i + ".");
	    inst = null;
	  }
	  state.setInstruction(a++, inst);
	}
      }
    }
    state.clearInstructionsFrom(a);
    String [] err;
    err = new String[errors.size()];
    for (i = 0; i < errors.size(); i++) {
      err[i] = (String)errors.elementAt(i);
    }
    return err;
  }

  int parseReg(StringTokenizer st, int line) throws NoSuchElementException {
    int result = parseWord(st, line, "register name");
    if (result > 31 || result < 0) {
      errors.addElement("Bad register number in line " + line + ".");
      result = 0;
    }
    return result;
  }

  int parseWord(StringTokenizer st, int line) throws NoSuchElementException {
    return parseWord(st, line, "label");
  }

  int parseWord(StringTokenizer st, int line, String labelType)
       throws NoSuchElementException {
    String r = st.nextToken();
    int result = 0;
    try {
      if (Character.isDigit(r.charAt(0)) ||
	  (r.length() > 1 && r.charAt(0) == '-' && Character.isDigit(r.charAt(1)))) {
	result = Integer.parseInt(r);
      } else {
	Integer lookup = (Integer)labels.get(r);
	if (lookup == null) {
          errors.addElement("Unknown name '" + r +  "' in line " + line +
                            ".");
        } else if(labelTypes.get(r) != labelType){
          errors.addElement("The name '" + r + "' is a " +
                            (String)(labelTypes.get(r)) + " but a " +
                            labelType + " is needed on line " + line + ".");
	} else {
	  result = lookup.intValue();
	}
      }
    }
    catch (NumberFormatException e) {
      errors.addElement("Bad word in line " + line + ".");
    }
    catch (StringIndexOutOfBoundsException e2) {
      errors.addElement("Internal error on line " + line + ".");
      System.err.println("In parseWord, st = '" + r + "': " + e2);
    }
    return result;
  }
}



